# Copyright (C) 2019 Maxim Godzi, Anatoly Zaytsev, Dmitrii Kiselev
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from sklearn.inspection import permutation_importance as perm_imp
from retentioneering.visualization import plot
import pandas as pd


class ModelDescriptor(object):

    def __init__(self, model, data, target, **kwargs):
        self.data = data
        self.target = target
        self.feature_extraction_kwargs = kwargs
        self.feature_extraction_kwargs.update({'ngram_range': kwargs.get('feature_range')})
        if hasattr(model, 'get_params'):
            model_filter = model.get_params(model)
            kwargs = {i: j for i, j in kwargs.items() if i in model_filter}

        try:
            self.mod = model(**kwargs)
        except:
            raise ValueError('Please use only keyword arguments from your model')
        self._fit_model()
        self.kwargs = kwargs

    def _fit_model(self):
        self.mod.fit(self.data, self.target)

    def permutation_importance_raw(self, test, **kwargs):
        test_sample = self.prepare_test(test)
        test_target = test.retention.get_positive_users()
        test_target = test_sample.index.isin(test_target)
        return self.permutation_importance(test_sample, test_target, node_params=None, **kwargs)

    def permutation_importance(self, test_sample, test_target, node_params=None, **kwargs):
        """
        Calculates permutation importance of features.
        If ``node_params`` is not `None`, then plots graph weighted by permutation importance.

        Parameters
        -------
        test_sample: pd.DataFrame
            Test feature subsample.
        test_target: np.array
            Vector of targets for test sample.
        node_params: dict, optional
            Event mapping describing which nodes or edges should be highlighted by different colors for better visualisation. Dictionary keys are ``event_col`` values, while keys have the following possible values:
                - ``bad_target``: highlights node and all incoming edges with red color;
                - ``nice_target``: highlights node and all incoming edges with green color;
                - ``bad_node``: highlights node with red color;
                - ``nice_node``: highlights node with green color;
                - ``source``: highlights node and all outgoing edges with yellow color.
            Example ``node_params`` is shown below:
            ```
            {
                'lost': 'bad_target',
                'purchased': 'nice_target',
                'onboarding_welcome_screen': 'source',
                'choose_login_type': 'nice_node',
                'accept_privacy_policy': 'bad_node',
            }
            ```
            If ``node_params=None``, it will be constructed from ``retention_config`` variable, so that:
            ```
            {
                'positive_target_event': 'nice_target',
                'negative_target_event': 'bad_target',
                'source_event': 'source',
            }
            ```
            Default: ``None``
        """
        self.show_quality_metrics(test_sample, test_target)
        perm = perm_imp(self.mod, test_sample, test_target, random_state=0)
        perm.pop('importances')
        perm['feature'] = test_sample.columns
        self.permutation_importance_table = pd.DataFrame(perm).sort_values('importances_mean', ascending=False)
        plot.permutation_importance(self.permutation_importance_table, **kwargs)
        self._plot_perm_imp(perm, test_sample, node_params, **kwargs)
        return self.permutation_importance_table

    def show_quality_raw(self, test):
        test_sample = self.prepare_test(test)
        test_target = test.retention.get_positive_users()
        test_target = test_sample.index.isin(test_target)
        return self.show_quality_metrics(test_sample, test_target, use_print=False)

    def show_quality_metrics(self, test_sample, test_target, use_print=True):
        """
        Calculate model quality metrics.

        Parameters
        -------
        test_sample: pd.DataFrame
            Test feature subsample.
        test_target: np.array
            Vector of targets for test sample.
        use_print: bool, optional
            If ``True``, prints out ROC-AUC, PR-AUC and accuracy in prediction task and RMSE, MAE and R-Squared in regression task.
        Returns
        -------
        ROC-AUC, PR-AUC and accuracy metrics in prediction task and RMSE, MAE and R-Squared in regression task.

        Return type
        -------
        Int
        """
        if hasattr(self.mod, 'predict_proba'):
            from sklearn.metrics import accuracy_score
            from sklearn.metrics import roc_auc_score
            from sklearn.metrics import average_precision_score
            preds = self.mod.predict_proba(test_sample)[:, 1]
            split = {}
            for i in range(1, 100):
                split.update({i: accuracy_score(test_target, preds > (i / 100))})
            best_split = pd.Series(split).idxmax() / 100
            roc = roc_auc_score(test_target, preds)
            aps = average_precision_score(test_target, preds)
            ac = accuracy_score(test_target, preds > best_split)
            if use_print:
                print(f"""
                ROC-AUC: {roc}
                PR-AUC: {aps}
                Accuracy: {ac}
                """)
            return roc, aps, ac
        else:
            from sklearn.metrics import mean_squared_error
            from sklearn.metrics import mean_absolute_error
            from sklearn.metrics import r2_score
            import numpy as np
            preds = self.mod.predict(test_sample)
            if use_print:
                print(f"""
                RMSE: {np.sqrt(mean_squared_error(test_target, preds))}
                MAE: {mean_absolute_error(test_target, preds)}
                R-squared: {r2_score(test_target, preds)}
                """)

    @staticmethod
    def _plot_perm_imp(perm, test_sample, node_params, **kwargs):
        weights = dict(zip(test_sample.columns.tolist(), perm["importances_mean"]))

        if node_params is None:
            node_params = {}
            node_flg = True
        else:
            node_flg = False
        node_weights = {}
        for node, val in weights.items():
            if len(node) > 1:
                continue
            if node_flg:
                node_params.update({
                    node[0]: 'nice_node' if weights[node] >= 0 else 'bad_node',
                })
            node_weights.update({
                node[0]: val
            })

        edge_cols = [i for i in test_sample.columns if len(i) == 2]
        if len(edge_cols) == 0:
            print("Sorry, you use only unigrams, change ngram_range to (1, 2) or greater")
            return
        data = []
        for key in edge_cols:
            data.append([key[0], key[1], weights.get(key)])

        plot.graph(pd.DataFrame(data), node_params, node_weights=node_weights, is_model=True, **kwargs)

    def visualize_results(self, plot_type='projections'):
        raise NotImplementedError('Sorry! This function is not ready now.')
        pass

    def predict(self, features):
        """
        Predicts probability of ``positive_target_event`` and ``negative_target_event`` in classifacation task or values of ``regeression_targets`` in regression task.

        Parameters
        -------
        features: pd.DataFrame
            Dataframe with test features. For more information refer to ``prepare_test()`` method.

        Returns
        -------
        Dataframe with predictions.

        Return type
        -------
        pd.DataFrame
        """
        if hasattr(self.mod, 'predict_proba'):
            return pd.DataFrame(self.mod.predict_proba(features), index=features.index, columns=[False, True])
        else:
            return pd.DataFrame(self.mod.predict(features), index=features.index, columns=['prediction'])

    def prepare_test(self, test):
        """
        Transforms test clickstream into train.

        Parameters
        ---------
        test: pd.DataFrame
            Raw clickstream.

        Returns
        --------
        Dataframe with test features.

        Return type
        -------
        pd.DataFrame
        """
        test = test.retention.extract_features(**self.feature_extraction_kwargs)
        test = test.loc[:, self.data.columns.tolist()]
        return test.fillna(0)

    def predict_raw(self, data):
        """
        Predicts probability of ``positive_target_event`` and ``negative_target_event`` in classifacation task or values of ``regeression_targets`` in regression task.

        Parameters
        --------
        data: pd.DataFrame
            Raw clickstream.

        Returns
        --------
        Dataframe with predictions.

        Return type
        --------
        pd.DataFrame
        """
        features = self.prepare_test(data)
        return self.predict(features)


class __LogRegWrapper__(object):
    def __init__(self, coef):
        self.coef = coef

    @property
    def feature_importances_(self):
        return self.coef



# Copyright (C) 2019 Maxim Godzi, Anatoly Zaytsev, Dmitrii Kiselev
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__VEGA_TEMPLATE__ = """
<html>
<head>
  <meta charset="UTF-8"
  <script src="https://api.retentioneering.com/files/d3.v4.min.js?a={func_name}"></script>
  <script type="text/javascript" src="https://cdn.jsdelivr.net/npm/vega@5"></script>
  <script type="text/javascript" src="https://cdn.jsdelivr.net/npm/vega-lite@4"></script>
  <script type="text/javascript" src="https://cdn.jsdelivr.net/npm/vega-embed@3.4"></script>

</head>
<body>
  <div id="visual-object-view"></div>

  <script type="text/javascript">
    var visualObject = {visual_object};

    vegaEmbed('#visual-object-view', visualObject);
  </script>
</body>
</html>
"""

__TEMPLATE__ = """
<html lang="en">
<head>
  <meta charset="UTF-8">
  <title>Graph Editor</title>
  <script src="https://code.jquery.com/jquery-3.4.1.js"></script>
  <script src="https://d3js.org/d3.v5.min.js"></script>
  <script type="text/javascript">

    var node_params;
    var mynodes = [];
    var mylinks = [];


    let maxDegree = 0;
    let maxWeigth = 0;
    let width = 800;
    let height = 800;


    function initialize(initNodes, initNodeParams, initLinks) {{
      mynodes = initNodes;
      node_params = initNodeParams;
      mylinks = initLinks;

      if (!{layout_dump}) {{
        let rawNodes = [];

        let delta = 0.1;
        for (i = 0; i < mynodes.length; i++) {{
          let newRawNode = null;

          //fixing correct positions for target nodes
          if (mynodes[i].type == 'bad_node') {{
            newRawNode = {{
              'index': mynodes[i].index,
              'name': mynodes[i].name,
              'type': mynodes[i].type,
              'degree': mynodes[i].degree,
              'fx': width * (1 - delta),
              'fy': height / 2
            }};
          }} else if (mynodes[i].type == 'nice_node') {{
            newRawNode = {{
              'index': mynodes[i].index,
              'name': mynodes[i].name,
              'type': mynodes[i].type,
              'degree': mynodes[i].degree,
              'fx': width * delta,
              'fy': height / 2
            }};
          }} else {{
            //for non-target nodes there x-s and y-s will be added after forceSimulation
            newRawNode = {{
              'index': mynodes[i].index,
              'name': mynodes[i].name,
              'type': mynodes[i].type,
              'degree': mynodes[i].degree
            }};
          }}
          //needed for normalization later
          if (mynodes[i].degree > maxDegree) {{
            maxDegree = mynodes[i].degree;
          }}

          rawNodes.push(newRawNode);
        }}
        var layout = d3
          .forceSimulation(rawNodes)
          .tick(5)
          ;


        let maxX = 0,
            maxY = 0,
            minX = 0,
            minY = 0;

        for (let i = 0; i < rawNodes.length; i++) {{
          if (rawNodes[i].type == 'suit_node') {{
            if (rawNodes[i].x > maxX) {{
              maxX = rawNodes[i].x;
            }}
            if (rawNodes[i].x < minX) {{
              minX = rawNodes[i].x;
            }}
            if (rawNodes[i].y > maxY) {{
              maxY = rawNodes[i].y;
            }}
            if (rawNodes[i].y < minY) {{
              minY = rawNodes[i].y;
            }}
          }}
        }}


        let offsetMaxX = -minX + maxX;
        let offsetMaxY = -minY + maxY;

        //Coordinates now have some unpredicted values. I set them so they fill my viewbox with some padding from borders.
        //At first I normalize them, then multiply by width and height.

        for (let i = 0; i < rawNodes.length; i++) {{

          if (rawNodes[i].type == 'suit_node') {{
            //x, y >= 0
            rawNodes[i].x += -minX;
            rawNodes[i].y += -minY;

            //x, y from [0, 1 - 2 * delta]
            rawNodes[i].x = rawNodes[i].x / offsetMaxX * (1 - 2 * delta);
            rawNodes[i].y = rawNodes[i].y / offsetMaxY * (1 - 2 * delta);

            //x, y from [delta, 1 - delta]
            rawNodes[i].x += delta;
            rawNodes[i].y += delta;

            //x, y from [(width | height) * delta, (width | height) * (1 - delta)]

            rawNodes[i].x *= width;
            rawNodes[i].y *= height;

          }}

        }}


        for (let i = 0; i < mynodes.length; i++) {{
          mynodes[i].x = rawNodes[i].x;
          mynodes[i].y = rawNodes[i].y;
        }}

        for (let i = 0; i < mylinks.length;  i++) {{
          mylinks[i].source.x = rawNodes[mylinks[i].source.index].x;
          mylinks[i].source.y = rawNodes[mylinks[i].source.index].y;
          mylinks[i].target.x = rawNodes[mylinks[i].target.index].x;
          mylinks[i].target.y = rawNodes[mylinks[i].target.index].y;
          if (mylinks[i].weight > maxWeigth) {{
            maxWeigth = mylinks[i].weight;
          }}
        }}

      }}

      makeCheckboxes();
      setLinkThreshold();
    }}

    function drawGraph(nodes, links) {{
      zoom = d3.zoom()
          .scaleExtent([0.5, 8])
          .translateExtent([[0, 0], [width, height]])
          .extent([[0, 0], [width, height]])
          .on("zoom", zoomed)
          ;

      var svg = d3.select("#freakingGraph").append("svg")
        .attr("viewBox", [0, 0, width, height])
        .call(zoom)
        ;

      //I append all elemets to maingroup so zoom works properly
      var maingroup = svg.append('g');

      function zoomed() {{
        maingroup.attr("transform", d3.event.transform);
      }}


      function calcMarkers(d) {{

          let dist = Math.sqrt((nodes[whereEquals(d.target.index)].x - nodes[whereEquals(d.source.index)].x) ** 2 + (nodes[whereEquals(d.target.index)].y - nodes[whereEquals(d.source.index)].y) ** 2);
          if (dist > 0 && dist <= 200){{
              return - Math.sqrt((0.5 - (d.target.degree ) / 2 / dist)) * (d.target.degree) / 2;

          }} else {{
              return 0;
          }}
      }}

      var path = maingroup.append("g").selectAll("path")
          .data(links)
          .enter()
          .append("path")
          .attr("class", function(d) {{ return "link " + d.type; }})
          .attr("stroke-width", function(d) {{ return Math.max(d.weight * 20, 1); }})
          .attr("id", function(d,i) {{ return "link_"+i; }})
          .attr("d", linkArc)
          ;

      let textMarkersSelection = maingroup.append("g").selectAll("text")
          .data(links)
          .enter();

      textMarkersSelection.append("text")
          .style("dominant-baseline", "central")
          .style("font-size", "13px")
          .append("textPath")
          .attr("xlink:href", function(d,i) {{ return "#link_"+i; }})
          .attr("startOffset", "35%")
          .text("➤")
          ;

      textMarkersSelection.append("text")
          .style("dominant-baseline", "central")
          .style("font-size", "13px")
          .append("textPath")
          .attr("xlink:href", function(d,i) {{ return "#link_"+i; }})
          .attr("startOffset", "65%")
          .text("➤")
          ;

      var edgetext = maingroup.append("g").selectAll("text")
          .data(links)
          .enter().append("text")
          .append("textPath")
          .attr("xlink:href",function(d,i){{return "#link_"+i;}})
          .style("text-anchor","middle")
          .attr("startOffset", "50%")
          .attr("id", function(d,i) {{ return "node_text"+i; }})
          ;

      function whereEquals(index) {{
        for (var i = 0; i < nodes.length; i++) {{
          if (index == nodes[i].index) {{
            return i;
          }}
        }}
      }}

      function displayingWeights() {{
        d3.selectAll("#show-weights").each(function(d) {{
          cb = d3.select(this);
          if (cb.property("checked")) {{
            edgetext = edgetext.text(function(d) {{
                if ($('#show-percents')[0].checked) {{
                    return Math.round(d['weight_text'] * 100) + "%";
                }} else {{
                    return Math.round(d['weight_text'] * 100) / 100;
                }}
            }})
          }} else {{
            edgetext = edgetext.text(function(d) {{ return ; }})
          }}


        }})
      }};



        d3.selectAll("#show-weights").on("change", displayingWeights);
        d3.selectAll("#show-percents").on("change", displayingWeights);

        function dragstarted(d) {{
          d3.select(this).raise().classed("active", true);
        }}

        function dragged(d) {{
          d3.select(this)
          .attr("cx", d.x = Math.min(Math.max(d3.event.x, 0), width))
          .attr("cy", d.y = Math.min(Math.max(d3.event.y, 0), height));
        }}

        function dragended(d) {{

          d3.select(this).classed("active", false);
          path = path.attr("d", linkArc);

          text = text
            .attr('x', function(d) {{ return d.x; }})
            .attr('y', function(d) {{ return d.y; }})
            ;
          defs.attr("refY", function(d) {{ return calcMarkers(d); }});
          defs.append("path")
            .attr("d", "M0,-5L10,0L0,5");
        }};

        var circle = maingroup.append("g").selectAll("circle")
            .data(nodes)
            .enter().append("circle")
            .attr("class", function(d) {{ return "circle " + d.type; }})
            .attr("r", function(d) {{ return d.degree; }})
            .attr('cx', function(d) {{ return d.x; }})
            .attr('cy', function(d) {{ return d.y; }})
            .style("cursor", "default")
            .call(d3.drag()
                .on("start", dragstarted)
                .on("drag", dragged)
                .on("end", dragended));


        var text = maingroup.append("g").selectAll("text")
          .data(nodes)
          .enter().append("text")
          .attr('x', function(d) {{ return d.x; }})
          .attr('y', function(d) {{ return d.y; }})
          .attr('id', function(d) {{ return "node-name" + d.index }})
          .attr('class', 'node-name')
          .text(function(d) {{ return d.name; }})
          ;

        function linkArc(d) {{
          var dx = nodes[whereEquals(d.target.index)].x - nodes[whereEquals(d.source.index)].x,
              dy = nodes[whereEquals(d.target.index)].y - nodes[whereEquals(d.source.index)].y,
              dr = dx * dx + dy * dy;
              dr = Math.sqrt(dr);
            if (dr > 200) {{
              dr *= 5
            }} else {{
              dr /= 2
            }};
            if (dr > 0) {{return "M" + nodes[whereEquals(d.source.index)].x + "," + nodes[whereEquals(d.source.index)].y + "A" + (dr * 1.1) + "," + (dr * 1.1) + " 0 0,1 " + nodes[whereEquals(d.target.index)].x + "," + nodes[whereEquals(d.target.index)].y;}}
            else {{return "M" + nodes[whereEquals(d.source.index)].x + "," + nodes[whereEquals(d.source.index)].y + "A" + 20 + "," + 20 + " 0 1,0 " + (nodes[whereEquals(d.target.index)].x + 0.1) + "," + (nodes[whereEquals(d.target.index)].y + 0.1);}}
        }}

        //synch with 'show names' and 'show weights' checkboxes
        displayingWeights();
        changeNamesVisibility(document.getElementById("show-names").checked)
    }}



    function changeLabel(curinput) {{
      document.getElementById('label' + curinput.id.substring(4)).innerHTML = curinput.value;
      mynodes[whereEquals1(curinput.id.substring(4))].name = curinput.value;
      $(curinput).attr('size', curinput.value.length + 2)
    }};

    function whereEquals1(index) {{
      for (var i = 0; i < mynodes.length; i++) {{
        if (index == mynodes[i].index) {{
          return i;
        }}
      }}
    }}

    function makeCheckboxes() {{

      for (var i = 0; i < mynodes.length; i++) {{

        var newDiv = document.createElement('div');
        newDiv.id = '#checkdiv' + mynodes[i].index;
        $( '#check-boxes' ).append(newDiv);

        var newCheckbox = document.createElement('input');
        newCheckbox.type = 'checkbox';
        newCheckbox.id = 'checkbox' + mynodes[i].index;
        newCheckbox.checked = true;
        newCheckbox.className = 'checkbox-class node-checkbox';
        $( newDiv ).append(newCheckbox);

        var newNameInput = document.createElement('input');
        newNameInput.id = 'name-input' + mynodes[i].index;
        newNameInput.type = 'text';
        newNameInput.value = mynodes[i].name;
        $( newNameInput ).attr('size', newNameInput.value.length + 3);

        $( newDiv ).append(newNameInput);
        $( newNameInput ).on('keypress', updateName);

      }}
    }}

    function updateName() {{

      document.getElementById('node-name' + this.id.substring(10)).innerHTML = this.value;
      mynodes[this.id.substring(10)].name = this.value;
      this.size = this.value.length + 3;

    }}

    function getCorrectLinks(newIdx) {{
      var newLinks = [];
      for (var i = 0; i < mylinks.length; i++) {{
        if (newIdx.includes(mylinks[i].source.index) && newIdx.includes(mylinks[i].target.index)) {{
          newLinks.push(mylinks[i]);
        }}
      }}
      return newLinks;
    }}
    function clearSVG() {{
      $( 'svg' ).remove();
      $( '.node-edit' ).each(function() {{
        this.remove();
      }});
    }}
    function changeNodes() {{
      var newNodes = [];
      var newIdx = [];

      $( '.node-checkbox' ).each(function(i, obj) {{

        if (this.checked) {{
          newNodes.push(mynodes[i]);
          newIdx.push(mynodes[i].index);
        }}

      }});

      var newLinks = getCorrectLinks(newIdx)


      clearSVG();
      drawGraph(newNodes, newLinks);
    }}

    function setLinkThreshold () {{
      let idxInLinks = new Array(mynodes.length).fill(false);
      let newLinks = [];

      let thresholdValue = $('#threshold-link-range').val();
      let blockDeleteTargets = $('#block-targets')[0].checked;

      for (let i = 0; i < mylinks.length; i++) {{
        if (mylinks[i].target.type == 'nice_node' || mylinks[i].source.type == 'nice_node' || mylinks[i].target.type == 'bad_node' || mylinks[i].source.type == 'bad_node') {{
          if (blockDeleteTargets) {{

            newLinks.push(mylinks[i]);
            idxInLinks[mylinks[i].target.index] = true;
            idxInLinks[mylinks[i].source.index] = true;
          }} else if (mylinks[i].weight / maxWeigth >= thresholdValue) {{
            newLinks.push(mylinks[i]);
            idxInLinks[mylinks[i].target.index] = true;
            idxInLinks[mylinks[i].source.index] = true;
          }}
        }} else if (mylinks[i].weight / maxWeigth >= thresholdValue) {{
          newLinks.push(mylinks[i]);
          idxInLinks[mylinks[i].target.index] = true;
          idxInLinks[mylinks[i].source.index] = true;
        }}
      }}

      let newNodes = [];
      for (let i = 0; i < mynodes.length; i++) {{
        if (idxInLinks[i]) {{
          newNodes.push(mynodes[i]);
          $('#checkbox' + mynodes[i].index).prop('checked', true);
        }} else {{
          $('#checkbox' + mynodes[i].index).prop('checked', false);
        }}
      }}
      clearSVG();
      drawGraph(newNodes, newLinks);
    }}

    function updateLinkThresholdText(val) {{
      document.getElementById('threshold-link-text').innerHTML = val;
    }}

    function setNodeThreshold() {{
      var newNodes = [];
      var newIdx = [];

      let thresholdValue = $('#threshold-node-range').val();
      let blockDeleteTargets = $('#block-targets')[0].checked;

      for (let i = 0; i < mynodes.length; i++) {{
        if (mynodes[i].type == 'bad_node' || mynodes[i].type == 'nice_node') {{
          if (blockDeleteTargets) {{
            newNodes.push(mynodes[i]);
            newIdx.push(mynodes[i].index);
            $('#checkbox' + mynodes[i].index).prop('checked', true);
          }} else if (mynodes[i].degree >= $('#threshold-node-range').val() * maxDegree) {{
            newNodes.push(mynodes[i]);
            newIdx.push(mynodes[i].index);
            $('#checkbox' + mynodes[i].index).prop('checked', true);
          }} else {{
            $('#checkbox' + mynodes[i].index).prop('checked', false);
          }}
        }} else if (mynodes[i].degree >= $('#threshold-node-range').val() * maxDegree) {{
          newNodes.push(mynodes[i]);
          newIdx.push(mynodes[i].index);
          $('#checkbox' + mynodes[i].index).prop('checked', true);
        }} else {{
          $('#checkbox' + mynodes[i].index).prop('checked', false);
        }}
      }}
      var newLinks = getCorrectLinks(newIdx);

      clearSVG();
      drawGraph(newNodes, newLinks);
    }}

    function updateNodeThresholdText(val) {{
      document.getElementById('threshold-node-text').innerHTML = val;
    }}


    function changeNamesVisibility(isHidden) {{

      if (isHidden) {{
        $('.node-name').each(
          function() {{
            $(this).show();
          }});
      }} else {{
        $('.node-name').each(
          function() {{
            $(this).hide();
          }});
      }}
    }}

    function downloadLayout() {{
        var a = document.createElement("a");
        var file = new Blob([JSON.stringify(mynodes)], {{type: "text/json;charset=utf-8"}});
        a.href = URL.createObjectURL(file);
        a.download = "node_params.json";
        console.log(1);
        a.click();

    }}

  </script>

  <style type="text/css">
      watermark {{
        width: 100%;
      }}
      watermark h3 {{
        width: 100%;
        text-align: center;
      }}
      html {{
        font-size: 10px;
      }}

      circle {{
        fill: #ccc;
        stroke: #333;
        stroke-width: 1.5px;
      }}

      .circle.source_node {{
        fill: #f3f310;
      }}

      .circle.nice_node {{
        fill: green;
      }}

      .circle.bad_node {{
        fill: red;
      }}

      .link {{
        fill: none;
        stroke: #666;
        stroke-opacity: 0.7;
      }}

      #nice_target {{
        fill: green;
      }}

      .link.nice_target {{
        stroke: green;
      }}

      #source {{
        fill: yellow;
      }}

      .link.source {{
        stroke: #f3f310;
      }}

      .link.positive {{
        stroke: green;
      }}

      .link.negative {{
        stroke: red;
      }}

      #source {{
        fill: orange;
      }}

      .link.source1 {{
        stroke: orange;
      }}

      #bad_target {{
        fill: red;
      }}

      .link.bad_target {{
        stroke: red;
      }}
      text {{
        font: 12px sans-serif;
        pointer-events: none;
        text-shadow: 0 1px 0 #fff, 1px 0 0 #fff, 0 -1px 0 #fff, -1px 0 0 #fff;
      }}

      main li {{
        display: inline;
      }}
      .graphlist {{
        list-style-type: none;

      }}
      .graphloader {{
        margin-top: 5%;
        margin-bottom: 5%;
      }}
      .graphloader input {{
        margin: auto;
      }}

      h1 {{
        text-align: center;
      }}

      .bottom-checkbox {{
        margin-right: 5%;
        display: inline;
      }}

      .checkbox-class {{
        margin-right: 3px;
      }}

      .node-edit {{
        position: relative;
        font-size: 12px;
        border: none;
        background-color: rgba(1,1,1,0);
        text-shadow: 0 1px 0 #fff, 1px 0 0 #fff, 0 -1px 0 #fff, -1px 0 0 #fff;

      }}

      .node-edit:focus {{
        background-color: #ddd;
      }}

      #option {{
        margin-left: 5px;
      }}

      #freakingGraph {{
        border: solid 2px black;
        /*position: relative;*/
      }}

      .container {{
        margin: 0!important;
        padding-right: 0!important;
        max-width: 1200px!important;
      }}
      .col-8 {{
        padding: 0px 4px 0px 2px!important;
      }}
      .col-4 {{
        padding-right: 0px!important;
      }}
      @media (max-width: 576px) {{
        form label {{
          font-size: 10px;
        }}
      }}


      @media (max-width: 768px) {{
        form label {{
          font-size: 0.8rem;
        }}
      }}


      @media (max-width: 992px) {{
        form label {{
          font-size: 1rem;
        }}
      }}

      @media (max-width: 1200px) {{
        form label {{
          font-size: 1rem;
        }}
      }}

      @media (min-width: 1201px) {{
        form label {{
          font-size: 1.4rem;
        }}
      }}

  </style>
  <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">
</head>
<body>



  <main>

    <div class="container">

        <div class="row">
          <div class="watermark" style="z-index: 1010; background-color: #FFF; width: 100%">
            <h3>Retentioneering</h3>
          </div>
          <div class="col-8">

            <div id="freakingGraph" style="z-index: 1000">
              <!-- graph will be appended here -->
            </div>
          </div>
          <div class="col-4" style="z-index: 1010; background-color: #FFF">
            <form>
              <div id="check-boxes">

              </div>
              <br>
              <input name="submit" value="Update nodes" style="width: 80%;" type="button" onclick="changeNodes()">

            </form>

            <br>
            <br>
            <div style="z-index: 1010; background-color: #FFF">
              <h6>Nodes Threshold</h6>
              <input id="threshold-node-range" name="threshold-node" type="range" min="0" max="1" step="0.01" value="0.05"
              oninput="updateNodeThresholdText(this.value)" onchange="updateNodeThresholdText(this.value)">
              <label id="threshold-node-text">0.05</label>
              <input type="button" value="Set threshold" onclick="setNodeThreshold()">
            </div>
            <br>
            <div>
              <h6>Links Threshold</h6>
              <input id="threshold-link-range" name="threshold" type="range" min="0" max="1" step="0.01" value={thresh}
              oninput="updateLinkThresholdText(this.value)" onchange="updateLinkThresholdText(this.value)">
              <label id="threshold-link-text">{thresh}</label>
              <input type="button" value="Set threshold" onclick="setLinkThreshold()">
            </div>
          </div>




          <div class="col-12" style="z-index: 1010; background-color: #FFF">

            <div class="weight-checkbox bottom-checkbox">
              <input type="checkbox" class="checkbox checkbox-class" value="weighted" id="show-weights"><label> Show weights </label>
            </div>

            <div class="percent-checkbox bottom-checkbox">
              <input type="checkbox" class="checkbox checkbox-class" id="show-percents"><label> Percents </label>
            </div>

            <div class="bottom-checkbox">
              <input type="checkbox" class="checkbox checkbox-class" checked id="show-names" onchange="changeNamesVisibility(this.checked)"><label> Show names</label>
            </div>

            <div class="bottom-checkbox">
              <input type="checkbox" class="checkbox checkbox-class" checked id="block-targets"><label>Block targets deletion</label>
            </div>
            <div id="option">
              <input name="downloadButton"
              type="button"
              value="download"
              onclick="downloadLayout()" />
            </div>
          </div>
      </div>




  </main>

  <script src="https://api.retentioneering.com/files/d3.v4.min.js"></script>

  <script type="text/javascript">

    initialize({nodes}, {node_params}, {links});

    if (!{show_percent}) {{
      $('.percent-checkbox').hide();
    }}

  </script>

  <script src="https://code.jquery.com/jquery-3.3.1.slim.min.js" integrity="sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo" crossorigin="anonymous"></script>
  <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js" integrity="sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1" crossorigin="anonymous"></script>
  <script src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js" integrity="sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM" crossorigin="anonymous"></script>
  <script src="https://code.jquery.com/jquery-3.4.1.js"></script>
</body>
</html>
"""

__OLD_TEMPLATE__ = """
<!DOCTYPE html>
<meta charset="utf-8">
<style>
                circle {{
                  fill: #ccc;
                  stroke: #333;
                  stroke-width: 1.5px;
                }}
                .circle.source_node {{
                  fill: #f3f310;
                }}
                .circle.nice_node {{
                  fill: green;
                }}
                .circle.bad_node {{
                  fill: red;
                }}
                .link {{
                  fill: none;
                  stroke: #666;
                  stroke-opacity: 0.7;
                }}
                #nice_target {{
                  fill: green;
                }}
                .link.nice_target {{
                  stroke: green;
                }}
                #source {{
                  fill: yellow;
                }}
                .link.source {{
                  stroke: #f3f310;
                }}

                .link.positive {{
                  stroke: green;
                }}

                .link.negative {{
                  stroke: red;
                }}
                #source {{
                  fill: orange;
                }}
                .link.source1 {{
                  stroke: orange;
                }}
                #bad_target {{
                  fill: red;
                }}
                .link.bad_target {{
                  stroke: red;
                }}
                text {{
                  font: 12px sans-serif;
                  pointer-events: none;
                  text-shadow: 0 1px 0 #fff, 1px 0 0 #fff, 0 -1px 0 #fff, -1px 0 0 #fff;
                }}
</style>
<body>
<script src="https://api.retentioneering.com/files/d3.v4.min.js"></script>
<div>
  <input type="checkbox" class="checkbox" value="weighted"><label> Show weights </label>
</div>
<div id="option">
    <input name="downloadButton"
           type="button"
           value="download"
           onclick="downloadLayout()" />
</div>
<script>
var links = {links};
var node_params = {node_params};
var nodes = {nodes};
var width = {width},
    height = {height};
var svg = d3.select("body").append("svg")
    .attr("width", width)
    .attr("height", height);
let defs = svg.append("g").selectAll("marker")
    .data(links)
  .enter().append("marker")
    .attr("id", function(d) {{ return d.source.index + '-' + d.target.index; }})
    .attr("viewBox", "0 -5 10 10")
    .attr("refX", function(d) {{
        if (d.target.name !== d.source.name) {{
            return 7 + d.target.degree;
        }} else {{
            return 0;
        }}
    }})
    .attr("refY", calcMarkers)
    .attr("markerWidth", 10)
    .attr("markerHeight", 10)
    .attr("markerUnits", "userSpaceOnUse")
    .attr("orient", "auto");
defs.append("path")
    .attr("d", "M0,-5L10,0L0,5");
function calcMarkers(d) {{
    let dist = Math.sqrt((nodes[d.target.index].x - nodes[d.source.index].x) ** 2 + (nodes[d.target.index].y - nodes[d.source.index].y) ** 2);
    if (dist > 0 && dist <= 200){{
        return - Math.sqrt((0.5 - (d.target.degree ) / 2 / dist)) * (d.target.degree) / 2;
    }} else {{
        return 0;
    }}
}}
var path = svg.append("g").selectAll("path")
    .data(links)
  .enter().append("path")
    .attr("class", function(d) {{ return "link " + d.type; }})
    .attr("stroke-width", function(d) {{ return Math.max(d.weight * 20, 1); }})
    .attr("marker-end", function(d) {{ return "url(#" + d.source.index + '-' + d.target.index + ")"; }})
    .attr("id", function(d,i) {{ return "link_"+i; }})
    .attr("d", linkArc)
    ;
var edgetext = svg.append("g").selectAll("text")
    .data(links)
   .enter().append("text")
   .append("textPath")
    .attr("xlink:href",function(d,i){{return "#link_"+i;}})
    .style("text-anchor","middle")
    .attr("startOffset", "50%")
    ;

function update() {{
    d3.selectAll(".checkbox").each(function(d) {{
        cb = d3.select(this);
        if (cb.property("checked")) {{
            edgetext = edgetext.text(function(d) {{
                if ({show_percent}) {{
                    return Math.round(d.weight_text * 100) / 100;
                }} else {{
                    return Math.round(d.weight_text * 100) + "%";
                }}
            }})
        }} else {{
            edgetext = edgetext.text(function(d) {{ return ; }})
        }}
    }})
}};
d3.selectAll(".checkbox").on("change",update);
function dragstarted(d) {{
  d3.select(this).raise().classed("active", true);
}}
function dragged(d) {{
  d3.select(this).attr("cx", d.x = d3.event.x).attr("cy", d.y = d3.event.y);
}}
function dragended(d) {{
  d3.select(this).classed("active", false);
  path = path.attr("d", linkArc);
  text = text
        .attr('x', function(d) {{ return d.x; }})
        .attr('y', function(d) {{ return d.y; }})
        ;
  defs = defs.attr("refY", calcMarkers);
  defs.append("path")
    .attr("d", "M0,-5L10,0L0,5");
}};
var circle = svg.append("g").selectAll("circle")
    .data(nodes)
  .enter().append("circle")
    .attr("class", function(d) {{ return "circle " + d.type; }})
    .attr("r", function(d) {{ return d.degree; }})
    .attr('cx', function(d) {{ return d.x; }})
    .attr('cy', function(d) {{ return d.y; }})
    .call(d3.drag()
        .on("start", dragstarted)
        .on("drag", dragged)
        .on("end", dragended));
var text = svg.append("g").selectAll("text")
    .data(nodes)
  .enter().append("text")
    .attr('x', function(d) {{ return d.x; }})
    .attr('y', function(d) {{ return d.y; }})
    .text(function(d) {{ return d.name; }});
function linkArc(d) {{
  var dx = nodes[d.target.index].x - nodes[d.source.index].x,
      dy = nodes[d.target.index].y - nodes[d.source.index].y,
      dr = dx * dx + dy * dy;
      dr = Math.sqrt(dr);
      if (dr > 200) {{
        dr *= 5
      }} else {{
        dr /= 2
      }};
      if (dr > 0) {{return "M" + nodes[d.source.index].x + "," + nodes[d.source.index].y + "A" + (dr * 1.1) + "," + (dr * 1.1) + " 0 0,1 " + nodes[d.target.index].x + "," + nodes[d.target.index].y;}}
      else {{return "M" + nodes[d.source.index].x + "," + nodes[d.source.index].y + "A" + 20 + "," + 20 + " 0 1,0 " + (nodes[d.target.index].x + 0.1) + "," + (nodes[d.target.index].y + 0.1);}}
}}
function downloadLayout() {{
    var a = document.createElement("a");
    var file = new Blob([JSON.stringify(nodes)], {{type: "text/json;charset=utf-8"}});
    a.href = URL.createObjectURL(file);
    a.download = "node_params.json";
    a.click();
}}
</script>

"""

__SANDBOX_TEMPLATE__ = """
<html>
<head>
  <title>Fuck me</title>
  <script src="https://d3js.org/d3.v5.min.js"></script>
</head>
<body>
  <div id="graph" style="width: 80%;"></div>
  <script type="text/javascript">
      let height = 600;
      let width = 600;
      data = func()
      const svg = d3.select("#graph").append("svg")
          .attr("viewBox", [0, 0, width, height]);

      const circle = svg.selectAll("circle")
        .data(data)
        .join("circle")
          .attr("transform", d => `translate(${d})`)
          .attr("r", 1.5);

      svg.call(d3.zoom()
          .extent([[0, 0], [width, height]])
          .scaleExtent([1, 8])
          .on("zoom", zoomed));

      function zoomed() {
        const {transform} = d3.event;
        //console.log(transform);
        //console.log(d3.event);
        circle.attr("transform", d => `translate(${transform.apply(d)})`);
      }





    function func(){
      const randomX = d3.randomNormal(width / 2, 80);
      const randomY = d3.randomNormal(height / 2, 80);
      return Array.from({length: 2000}, () => [randomX(), randomY()]);
    }

  </script>
</body>
</html>

"""



# Copyright (C) 2019 Maxim Godzi, Anatoly Zaytsev, Dmitrii Kiselev
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import networkx as nx
import seaborn as sns
from IPython.display import IFrame, display  # TODO understand how to use visualization without it
import numpy as np
from datetime import datetime
import pandas as pd
import json
from functools import wraps
from retentioneering.visualization.cloud_logger import MongoLoader
import altair as alt
import os
from retentioneering.visualization import templates


def _calc_layout(data, node_params, width=500, height=500, **kwargs):
    G = nx.DiGraph()
    G.add_weighted_edges_from(data.loc[:, ['source', 'target', 'weight']].values)
    if nx.algorithms.check_planarity(G)[0]:
        pos_new = nx.layout.planar_layout(G)
    else:
        pos_new = nx.layout.spring_layout(G, k=kwargs.get('k', .1),
                                          iterations=kwargs.get('iterations', 300),
                                          threshold=kwargs.get('nx_threshold', 1e-4),
                                          seed=0)
    min_x = min([j[0] for i, j in pos_new.items()])
    min_y = min([j[1] for i, j in pos_new.items()])
    max_x = max([j[0] for i, j in pos_new.items()])
    max_y = max([j[1] for i, j in pos_new.items()])
    pos_new = {
        i: [(j[0] - min_x) / (max_x - min_x) * (width - 150) + 75,
            (j[1] - min_y) / (max_y - min_y) * (height - 100) + 50]
        for i, j in pos_new.items()
    }
    return pos_new, dict(G.degree)


def _prepare_nodes(data, pos, node_params, degrees):
    node_set = set(data['source']) | set(data['target'])
    max_degree = max(degrees.values())
    nodes = {}
    for idx, node in enumerate(node_set):
        node_pos = pos.get(node)
        nodes.update({node: {
            "index": idx,
            "name": node,
            "x": node_pos[0],
            "y": node_pos[1],
            "type": (node_params.get(node) or "suit").split('_')[0] + '_node',
            "degree": (abs(degrees.get(node, 0))) / abs(max_degree) * 30 + 4
        }})
    return nodes


def _prepare_edges(data, nodes):
    edges = []
    data['weight_norm'] = data.weight / data.weight.abs().max()
    for idx, row in data.iterrows():
        edges.append({
            "source": nodes.get(row.source),
            "target": nodes.get(row.target),
            "weight": np.log1p(abs(row.weight_norm + 1e-100)) * 1.5,
            "weight_text": row.weight,
            "type": row['type']
        })
    return edges, list(nodes.values())


def _filter_edgelist(data, thresh, node_params, targets=None, **kwargs):
    if targets is None:
        x = pd.Series(node_params).str.contains('target')
        targets = set(x[x].index)
    f = data.weight.abs() >= thresh
    nodes = set(data[f].source) | set(data[f].target)
    f |= (data.source.isin(targets) & data.target.isin(nodes))
    f |= (data.target.isin(targets) & data.source.isin(nodes))
    return data[f].copy()


def _make_json_data(data, node_params, layout_dump, thresh=.05, width=500, height=500, **kwargs):
    res = {}
    data.columns = ['source', 'target', 'weight']
    data = _filter_edgelist(data, thresh, node_params, **kwargs)
    if kwargs.get('mode') == 'importance':
        data['type'] = np.where(data.weight >= 0, 'positive', 'negative')
    else:
        data["type"] = data.apply(
            lambda x: node_params.get(x.source) if node_params.get(x.source) == 'source' else node_params.get(
                x.target) or 'suit',
            1
        )
    pos, degrees = _calc_layout(data, node_params, width=width, height=height, **kwargs)
    if kwargs.get('node_weights') is not None:
        degrees = kwargs.get('node_weights')
    if layout_dump is not None:
        nodes = _prepare_given_layout(layout_dump, node_params, degrees)
    else:
        nodes = _prepare_nodes(data, pos, node_params, degrees)
    res['links'], res['nodes'] = _prepare_edges(data, nodes)
    return res


def _prepare_node_params(node_params, data):
    if node_params is None:
        _node_params = {
            'positive_target_event': 'nice_target',
            'negative_target_event': 'bad_target',
            'source_event': 'source',
        }
        node_params = {}
        for key, val in _node_params.items():
            name = data.retention.retention_config.get(key)
            if name is None:
                continue
            node_params.update({name: val})
    return node_params


def _prepare_layout(layout):
    nodes = {}
    for i in layout:
        nodes.update({i['name']: i})
    return nodes


def _prepare_given_layout(nodes_path, node_params, degrees):
    if type(nodes_path) is str:
        with open(nodes_path, encoding='utf-8') as f:
            nodes = json.load(f)
    else:
        nodes = nodes_path
    if type(nodes) is list:
        nodes = _prepare_layout(nodes)
    max_degree = max(degrees.values() or [1e-20])
    for node, val in nodes.items():
        val.update({
            "type": (node_params.get(node) or "suit").split('_')[0] + '_node',
            "degree": degrees.get(node, 0) / max_degree * 30
        })
    return nodes


def __save_plot__(func):
    @wraps(func)
    def save_plot_wrapper(*args, **kwargs):
        sns.mpl.pyplot.show()
        sns.mpl.pyplot.close()
        res = func(*args, **kwargs)
        if len(res) == 2:
            (vis_object, name), res, cfg = res, None, None
        elif len(res) == 3:
            (vis_object, name, res), cfg = res, None
        else:
            vis_object, name, res, cfg = res
        idx = 'id: ' + str(int(datetime.now().timestamp()))
        coords = vis_object.axis()


        if '_3d_' not in name:
            vis_object.text((coords[0] - (coords[1] - coords[0]) / 10),
                            (coords[3] + (coords[3] - coords[2]) / 10), idx, fontsize=8)
            vis_object.text(0.25, 0.05, 'Retentioneering', fontsize=50, color='gray', va='bottom', alpha=0.1)
        vis_object.get_figure().savefig(name, bbox_inches="tight", dpi=cfg.get('save_dpi') or 200)
        if cfg.get('mongo_client') is not None:
            print(f'DB {idx}')
            ml = MongoLoader(cfg.get('mongo_client'), collection=cfg.get('mongo_user'))
            ml.put(name if '.' in name else name + '.png', idx.split(' ')[1])
            if '.html' in name:
                ml.put(vis_object.get_raw(name), idx.split(' ')[1] + '_config')
        return res
    return save_plot_wrapper


def __altair_save_plot__(func):
    @wraps(func)
    def altair_save_plot_wrapper(*args, **kwargs):
        res = func(*args, **kwargs)
        vis_object, plot_name, res, cfg = res
        idx = 'id: ' + str(int(datetime.now().timestamp())) + ", Retentioneering Copyright"
        plot_name_preffix = func.__name__

        plot_name = '{}_{}'.format(plot_name_preffix, plot_name or datetime.now()).replace(':', '_').replace('.',
                                                                                                             '_').replace(
            ' ', '_') + '.html'
        if hasattr(pd.DataFrame, 'retention'):
            plot_name = pd.DataFrame().retention.retention_config['experiments_folder'] + '/' + plot_name
        else:
            path = 'experiments'
            if not os.path.exists(path):
                os.mkdir(path)
            plot_name = path + '/' + plot_name
        vis_object.title = idx

        print("You can save plot as SVG or PNG by open three-dotted button at right =>")
        watermark = alt.Chart().mark_text(
            align='center', baseline='top', dy=vis_object.height // 2 + 30, fontSize=32, fontWeight=200,
            color='#d3d3d3', text='Retentioneering'
        )
        vis_object.save(plot_name)

        render_in_template = True

        if render_in_template and kwargs.get('interactive', True):
            html_object = templates.__VEGA_TEMPLATE__.format(
                visual_object=json.dumps(vis_object.to_dict()),
                func_name=func.__name__
            )
            fig = ___DynamicFigureWrapper__(html_object, True, vis_object.width, vis_object.height, res)
            fig.get_figure().savefig(plot_name, bbox_inches="tight", dpi=cfg.get('save_dpi') or 200)

        elif kwargs.get('interactive', True):
            print("You can save plot as SVG or PNG by open three-dotted button at right =>")
            alt.renderers.enable('notebook')
            display(vis_object)

        if cfg.get('mongo_client') is not None:
            print(f'DB {idx}')
            ml = MongoLoader(cfg.get('mongo_client'), collection=cfg.get('mongo_user'))
            ml.put(plot_name if '.' in plot_name else plot_name + '.svg', idx.split(' ')[1])
        return res
    return altair_save_plot_wrapper


@__save_plot__
def graph(data, node_params=None, thresh=.05, width=800, height=500, interactive=True,
          layout_dump=None, show_percent=True, plot_name=None, node_weights=None, **kwargs):
    """
    Create interactive graph visualization. Each node is a unique ``event_col`` value, edges are transitions between events and edge weights are calculated metrics. By default, it is a percentage of unique users that have passed though a particular edge visualized with the edge thickness. Node sizes are  Graph loop is a transition to the same node, which may happen if users encountered multiple errors or made any action at least twice.
    Graph nodes are movable on canvas which helps to visualize user trajectories but is also a cumbersome process to place all the nodes so it forms a story.
    That is why IFrame object also has a download button. By pressing it, a JSON configuration file with all the node parameters is downloaded. It contains node names, their positions, relative sizes and types. It it used as ``layout_dump`` parameter for layout configuration. Finally, show weights toggle shows and hides edge weights.

    Parameters
    ---------
    data: pd.DataFrame
        Graph in edgelist form.
    node_params: dict, optional
        Event mapping describing which nodes or edges should be highlighted by different colors for better visualisation. Dictionary keys are ``event_col`` values, while keys have the following possible values:
            - ``bad_target``: highlights node and all incoming edges with red color;
            - ``nice_target``: highlights node and all incoming edges with green color;
            - ``bad_node``: highlights node with red color;
            - ``nice_node``: highlights node with green color;
            - ``source``: highlights node and all outgoing edges with yellow color.
        Example ``node_params`` is shown below:
        ```
        {
            'lost': 'bad_target',
            'purchased': 'nice_target',
            'onboarding_welcome_screen': 'source',
            'choose_login_type': 'nice_node',
            'accept_privacy_policy': 'bad_node',
        }
        ```
        If ``node_params=None``, it will be constructed from ``retention_config`` variable, so that:
        ```
        {
            'positive_target_event': 'nice_target',
            'negative_target_event': 'bad_target',
            'source_event': 'source',
        }
        ```
    thresh: float, optional
        Minimal edge weight value to be rendered on a graph. If a node has no edges of the weight >= ``thresh``, then it is not shown on a graph. It is used to filter out rare event and not to clutter visualization. Default: ``0.05``
    width: float, optional
        Width of plot in pixels. Default: ``500``
    height: float, optional
        Height of plot in pixels. Default: ``500``
    interactive: bool, optional
        If ``True``, then plots graph visualization in interactive session (Jupyter notebook). Default: ``True``
    layout_dump: str, optional
        Path to layout configuration file relative to current directory. If defined, uses configuration file as a graph layout. Default: ``None``
    show_percent: bool, optional
        If ``True``, then all edge weights are converted to percents by multiplying by 100 and adding percentage sign. Default: ``True``

    Returns
    -------
    Saves webpage with JS graph visualization to ``retention_config.experiments_folder``.

    Return type
    -------
    HTML
    """
    dump = 1
    if layout_dump is None:
        dump = 0
    if node_params is None:
        node_params = _prepare_node_params(node_params, data)
    res = _make_json_data(data, node_params, layout_dump, thresh=thresh,
                          width=width - width / 3, height=height - height / 3, node_weights=node_weights, **kwargs)
    res['node_params'] = node_params
    show = 0
    if show_percent:
        show = 1
    dump = 1 if (layout_dump is not None) or (kwargs.get('is_model', False)) else 0
    __TEMPLATE__ = templates.__OLD_TEMPLATE__ if kwargs.get('use_old', False) else templates.__TEMPLATE__
    x = __TEMPLATE__.format(
        width=width,
        height=height,
        links=json.dumps(res.get('links')).encode('latin1').decode('utf-8'),
        node_params=json.dumps(node_params).encode('latin1').decode('utf-8'),
        nodes=json.dumps(res.get('nodes')).encode('latin1').decode('utf-8'),
        show_percent=show,
        layout_dump=dump,
        thresh=thresh,
    )
    if hasattr(data, 'trajectory'):
        if plot_name is None:
            plot_name = f'index_{datetime.now()}'
    else:
        if plot_name is None:
            plot_name = 'index'
    plot_name = f"{data.trajectory.retention_config['experiments_folder']}/{plot_name.replace(':', '_').replace('.', '_')}" + '.html'
    return (
        ___DynamicFigureWrapper__(x, interactive, width, height, res),
        plot_name,
        plot_name,
        data.retention.retention_config
    )


@__altair_save_plot__
def altair_step_matrix(diff, plot_name=None, title='', vmin=None, vmax=None, font_size=12, **kwargs):
    heatmap_data = diff.reset_index().melt('index')
    heatmap_data.columns = ['y', 'x', 'z']
    table = alt.Chart(heatmap_data).encode(
        x=alt.X('x:O', sort=None),
        y=alt.Y('y:O', sort=None)
    )
    heatmap = table.mark_rect().encode(
        color=alt.Color(
            'z:Q',
            scale=alt.Scale(scheme='blues'),
        )
    )
    text = table.mark_text(
        align='center', fontSize=font_size
    ).encode(
        text='z',
        color=alt.condition(
            abs(alt.datum.z) < 0.8,
            alt.value('black'),
            alt.value('white'))
    )
    heatmap_object = (heatmap + text).properties(
        width=3 * font_size * len(diff.columns),
        height=2 * font_size * diff.shape[0]
    )
    return heatmap_object, plot_name, None, diff.retention.retention_config


@__save_plot__
def step_matrix(diff, plot_name=None, title='', vmin=None, vmax=None, **kwargs):
    """
    Plots heatmap with distribution of users over events and trajectories steps ordered by timestamp. Rows are event names, columns are trajectory steps (in session or user story) and values are shares of users experiencing specific event on a specific step of trajectory. If used with `diff` filter, then shows users share difference between datasets.

    Parameters
    --------
    diff: pd.DataFrame
        Table for heatmap visualization.
    plot_name: str, optional
        Name of the plot to save. Default: ``None``
    title: str, optional
        Title for step matrix plot.
    kwargs: optional
        ``sns.heatmap`` parameters.

    Returns
    --------
    Dataframe with ``max_steps`` number of columns and ``len(event_col.unique)`` number of rows at max, or less if used ``thr`` > 0.
    Saves heatmap to ``retention_config.experiments_folder``

    Return type
    --------
    pd.DataFrame
    PNG
    """

    sns.mpl.pyplot.figure(figsize=(20, 10))
    heatmap = sns.heatmap(diff, annot=True, cmap="BrBG", center=0, vmin=vmin, vmax=vmax)
    heatmap.set_title(title)
    plot_name = plot_name or 'step_matrix_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = diff.retention.retention_config['experiments_folder'] + '/' + plot_name
    return heatmap, plot_name, None, diff.retention.retention_config


@__altair_save_plot__
def altair_cluster_tsne(data, clusters, target, plot_name=None, **kwargs):

    if hasattr(data.retention, '_tsne'):
        tsne = data.retention._tsne.copy()
    else:
        tsne = data.retention.learn_tsne(clusters, **kwargs)
    tsne['color'] = clusters
    tsne.columns = ['x', 'y', 'color']

    scatter = alt.Chart(tsne).mark_point().encode(
        x='x',
        y='y',
        color=alt.Color(
            'color',
            scale=alt.Scale(scheme='plasma')
        )
    ).properties(
        width=800,
        height=600
    )
    return scatter, plot_name, tsne, data.retention.retention_config


@__save_plot__
def cluster_tsne(data, clusters, target, plot_name=None, **kwargs):
    """
    Plots TSNE projection of user stories colored by clusters. Each point represents a user session or whole user trajectory.

    Parameters
    --------
    data: pd.DataFrame
        Feature matrix.
    clusters: np.array
        Array of cluster IDs.
    target: np.array
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.
    plot_name: str, optional
        Name of plot to save. Default: ``'clusters_tsne_{timestamp}.svg'``

    Returns
    -------
    Saves plot to ``retention_config.experiments_folder``

    Return type
    -------
    PNG
    """

    if hasattr(data.retention, '_tsne'):
        tsne2 = data.retention._tsne.copy()
    else:
        tsne2 = data.retention.learn_tsne(clusters, **kwargs)
    tsne = tsne2.values
    if np.unique(clusters).shape[0] > 10:
        f, ax = sns.mpl.pyplot.subplots()
        points = ax.scatter(tsne[:, 0], tsne[:, 1], c=clusters, cmap="BrBG")
        f.colorbar(points)
        scatter = ___FigureWrapper__(f)
    else:
        scatter = sns.scatterplot(tsne[:, 0], tsne[:, 1], hue=clusters, legend='full', palette= sns.color_palette("bright")[0:np.unique(clusters).shape[0]])
    plot_name = plot_name or 'cluster_tsne_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.retention.retention_config['experiments_folder'] + '/' + plot_name
    return scatter, plot_name, tsne2, data.retention.retention_config

@__altair_save_plot__
def altair_cluster_bar(data, clusters, target, plot_name=None, plot_cnt=None, metrics=None, **kwargs):
    cl = pd.DataFrame([clusters, target], index=['clusters', 'target']).T
    cl['cnt'] = 1
    cl.target = cl.target.astype(int)
    bars = cl.groupby('clusters').agg({
        'cnt': 'sum',
        'target': 'mean'
    }).reset_index()
    bars.cnt /= bars.cnt.sum()
    bars = bars.loc[:, ['clusters', 'cnt']].append(bars.loc[:, ['clusters', 'target']], ignore_index=True, sort=False)
    bars['target'] = np.where(bars.target.isnull(), bars.cnt, bars.target)
    bars['Metric'] = np.where(bars['cnt'].isnull(), 'Average CR', 'Cluster size')
    # print(bars, type(bars))
    bar = alt.Chart(bars).mark_bar().encode(
        x='Metric:O',
        y='target:Q',
        color='Metric:N',
        column='clusters:N'
    ).properties(
        width=60,
        height=200
    )

    return bar, plot_name, None, data.retention.retention_config






@__save_plot__
def cluster_bar(data, clusters, target, plot_name=None, plot_cnt=None, metrics=None, **kwargs):
    """
    Plots bar charts with cluster sizes and average target conversion rate.

    Parameters
    ----------
    data : pd.DataFrame
        Feature matrix.
    clusters : np.array
        Array of cluster IDs.
    target: np.array
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.
    plot_name : str, optional
        Name of plot to save. Default: ``'clusters_bar_{timestamp}.svg'``
    kwargs: optional
        Width and height of plot.

    Returns
    -------
    Saves plot to ``retention_config.experiments_folder``

    Return type
    -------
    PNG
    """
    cl = pd.DataFrame([clusters, target], index=['clusters', 'target']).T
    cl['cnt'] = 1
    cl.target = cl.target.astype(int)
    bars = cl.groupby('clusters').agg({
        'cnt': 'sum',
        'target': 'mean'
    }).reset_index()
    bars.cnt /= bars.cnt.sum()
    bars = bars.loc[:, ['clusters', 'cnt']].append(bars.loc[:, ['clusters', 'target']], ignore_index=True, sort=False)
    bars['target'] = np.where(bars.target.isnull(), bars.cnt, bars.target)
    bars['Metric'] = np.where(bars['cnt'].isnull(), 'Average CR', 'Cluster size')
    bar = sns.barplot(x='clusters', y='target', hue='Metric', hue_order=['Cluster size', 'Average CR'], data=bars)
    y_value = ['{:,.2f}'.format(x * 100) + '%' for x in bar.get_yticks()]
    bar.set_yticklabels(y_value)
    bar.set(ylabel=None)

    plot_name = plot_name or 'cluster_bar_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.retention.retention_config['experiments_folder'] + '/' + plot_name
    return bar, plot_name, None, data.retention.retention_config


@__save_plot__
def cluster_event_dist(bars, event_col, cl1, sizes, crs, cl2=None, plot_name=None):
    bar = sns.barplot(x=event_col, y='freq', hue='hue',
                      hue_order=[f'cluster {cl1}','all' if cl2 is None else f'cluster {cl2}'], data=bars)
    y_value = ['{:,.2f}'.format(x * 100) + '%' for x in bar.get_yticks()]
    bar.set_yticklabels(y_value)
    bar.set_xticklabels(bar.get_xticklabels(), rotation=30)
    bar.set(ylabel=None)
    tit = f'Distribution of top {bars.shape[0] // 2} events in cluster {cl1} (size: {round(sizes[0] * 100, 2)}%, CR: {round(crs[0] * 100, 2)}% ) '
    tit += f'vs. all data (CR: {round(crs[1] * 100, 2)}%)' if cl2 is None else f'vs. cluster {cl2} (size: {round(sizes[1] * 100, 2)}%, CR: {round(crs[1] * 100, 2)}%)'
    bar.set_title(tit)

    plot_name = plot_name or 'cluster_event_dist_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = bars.retention.retention_config['experiments_folder'] + '/' + plot_name
    return bar, plot_name, None, bars.retention.retention_config


@__save_plot__
def cluster_pie(data, clusters, target, plot_name=None, plot_cnt=None, metrics=None, **kwargs):
    """
    Plots pie-charts of target distribution for different clusters.

    Parameters
    -------
    data: pd.DataFrame
        Feature matrix
    clusters: np.array
        Array of cluster IDs.
    target: np.array
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.
    plot_name: str, optional
        Name of plot to save. Default: ``'clusters_pie_{timestamp}.svg'``
    kwargs: optional
        Width and height of plot.

    Returns
    -------
    Saves plot to ``retention_config.experiments_folder``

    Return type
    -------
    PNG
    """
    cl = pd.DataFrame([clusters, target], index=['clusters', 'target']).T
    cl.target = np.where(cl.target, data.retention.retention_config['positive_target_event'],
                         data.retention.retention_config['negative_target_event'])
    pie_data = cl.groupby(['clusters', 'target']).size().rename('target_dist').reset_index()
    targets = list(set(pie_data.target))

    if plot_cnt is None:
        plot_cnt = len(set(clusters))

    if kwargs.get('vol', True):  # vol = False in kwargs in case you want to disable
        _, counts = np.unique(clusters, return_counts=True)
        volumes = 100 * (counts / sum(counts))
    else:
        volumes = [None] * plot_cnt

    fig, ax = sns.mpl.pyplot.subplots(1 if plot_cnt <= 2 else (plot_cnt // 2 + plot_cnt % 2), 2)
    fig.suptitle(
        'Distribution of targets in clusters. Silhouette: {:.2f}, Homogeneity: {:.2f}, Cluster stability: {:.2f}'.format(
            metrics.get('silhouette') if (metrics or {}).get('silhouette') is not None else 0,
            metrics.get('homogen') if metrics is not None else 0,
            metrics.get('csi') if (metrics or {}).get('csi') is not None else 0
        ))
    fig.set_size_inches(kwargs.get('width', 20), kwargs.get('height', 10))
    for i, j in enumerate(pie_data.clusters.unique()):
        tmp = pie_data[pie_data.clusters == j]
        tmp.index = tmp.target
        if plot_cnt <= 2:
            ax[i].pie(tmp.target_dist.reindex(targets).fillna(0).values, labels=targets, autopct='%1.1f%%')
            ax[i].set_title('Class {}\nCluster volume {}%\nMean dist from center {:.2f}'.format(
                i, round(volumes[i], 1), metrics['mean_fc'][j] if (metrics or {}).get('mean_fc') is not None else 0))
        else:
            ax[i // 2][i % 2].pie(tmp.target_dist.reindex(targets).fillna(0).values, labels=targets, autopct='%1.1f%%')
            ax[i // 2][i % 2].set_title('Class {}\nCluster volume {}%\nMean dist from center {:.2f}'.format(
                i, round(volumes[i], 1), metrics['mean_fc'][j] if (metrics or {}).get('mean_fc') is not None else 0))
    if plot_cnt % 2 == 1:
        fig.delaxes(ax[plot_cnt // 2, 1])

    plot_name = plot_name or 'cluster_pie_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.retention.retention_config['experiments_folder'] + '/' + plot_name
    return ___FigureWrapper__(fig), plot_name, None, data.retention.retention_config


@__save_plot__
def cluster_heatmap(data, clusters, target, plot_name=None, **kwargs):
    """
    Visualizes feature usage with heatmap.

    Parameters
    --------
    data: pd.DataFrame
        Feature matrix.
    clusters: np.array
        Array of cluster IDs.
    target: np.array
        Boolean vector, if ``True``, then user has `positive_target_event` in trajectory.
    plot_name: str, optional
        Name of plot to save. Default: ``'clusters_heatmap_{timestamp}.svg'``

    Returns
    -------
    Saves plot to ``retention_config.experiments_folder``

    Return type
    -------
    PNG
    """
    heatmap = sns.clustermap(data.values,
                             cmap="BrBG",
                             xticklabels=data.columns,
                             yticklabels=False,
                             row_cluster=True,
                             col_cluster=False)

    heatmap.ax_row_dendrogram.set_visible(False)
    heatmap = heatmap.ax_heatmap

    plot_name = plot_name or 'cluster_heatmap_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.retention.retention_config['experiments_folder'] + '/' + plot_name
    return heatmap, plot_name, None, data.retention.retention_config


@__save_plot__
def core_event_dist(rates, thresh, plot_name=None, **kwargs):
    hist = sns.distplot(rates.values, hist=True, bins=kwargs.get('bins'), kde=kwargs.get('kde'))
    if thresh is not None:
        sns.mpl.pyplot.axvline(thresh, c='C1')

    plot_name = plot_name or 'core_event_dist_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    rates = rates.reset_index()
    plot_name = rates.retention.retention_config['experiments_folder'] + '/' + plot_name
    return hist, plot_name, None, rates.retention.retention_config


@__save_plot__
def permutation_importance(x, plot_name=None, **kwargs):
    fig = sns.mpl.pyplot.figure(figsize=[20, 7])
    sns.mpl.pyplot.xticks(rotation="vertical")
    sns.mpl.pyplot.title("Permutation importances")
    sns.mpl.pyplot.bar(x.feature.map(lambda x: " ".join(x)), x.importances_mean, yerr=x.importances_std)
    plot_name = plot_name or 'permutation_importance_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    return ___FigureWrapper__(fig), plot_name, None, x.retention.retention_config

@__save_plot__
def tsne_3d(data, clusters, target, plot_name=None, use_coloring=False, **kwargs):
    if hasattr(data.retention, '_tsne'):
        tsne2 = data.retention._tsne.copy()
    else:
        tsne2 = data.retention.learn_tsne(clusters, **kwargs)
    tsne = tsne2.values
    fig = sns.mpl.pyplot.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')

    if target is not None:
        scatter = ax.scatter(tsne[:, 0], tsne[:, 1], target, c=['C{}'.format(i) for i in clusters])
        lgs = []
        for i in set(clusters):
            lgs.append(sns.mpl.lines.Line2D([0], [0], linestyle="none", c='C{}'.format(i), marker='o'))
        ax.legend(lgs, set(clusters), numpoints=1)
    else:
        scatter = ax.scatter(tsne[:, 0], tsne[:, 1], clusters)

    ax.set_xlabel('TSNE 0')
    ax.set_ylabel('TSNE 1')
    ax.set_zlabel('Target')

    scatter = ___FigureWrapper__(fig)
    plot_name = plot_name or 'tsne_3d_{}'.format(datetime.now()).replace(':', '_').replace('.', '_') + '.svg'
    plot_name = data.retention.retention_config['experiments_folder'] + '/' + plot_name
    return scatter, plot_name, None, data.retention.retention_config


class ___FigureWrapper__(object):
    def __init__(self, fig):
        self.fig = fig

    def get_figure(self):
        return self.fig

    def axis(self):
        if len(self.fig.axes) > 1:
            x = self.fig.axes[1].axis()
        else:
            x = self.fig.axes[0].axis()
        return (x[0]/ 64, x[0] + (x[1] - x[0]) / 50, x[2] / 1.5, x[3] / 1.5)

    def text(self, *args, **kwargs):
        self.fig.text(*args, **kwargs)


class __SaveFigWrapper__(object):
    def __init__(self, data, interactive=True, width=1000, height=700):
        self.data = data
        self.interactive = interactive
        self.width = width
        self.height = height

    def savefig(self, name, **kwargs):
        with open(name, 'w', encoding="utf-8") as f:
            f.write(self.data)
        if self.interactive:
            display(IFrame(name, width=self.width + 200, height=self.height + 200))


class ___DynamicFigureWrapper__(object):
    def __init__(self, fig, interactive, width, height, links):
        self.fig = fig
        self.interactive, self.width, self.height = interactive, width, height
        self.links = links

    def get_figure(self):
        savefig = __SaveFigWrapper__(self.fig, self.interactive, self.width, self.height)
        return savefig

    def text(self, x, y, text, *args, **kwargs):
        # parts = self.fig.split('<main>')
        # res = parts[:1] + [f'<p>{text}</p>'] + parts[1:]
        # self.fig = '\n'.join(res)
        pass

    def get_raw(self, path):
        base = '.'.join(path.split('.')[:-1])
        with open(base + '_config.json', 'w') as f:
            json.dump(self.links, f)
        return base + '_config.json'

    @staticmethod
    def axis():
        return 4 * [0]



from setuptools import setup, find_packages

setup(
    name='retentioneering',
    version='1.0.7.5',
    license='Mozilla Public License',
    description='Product analytics and marketing optimization framework based on deep user trajectories analysis',
    long_description="""
    Retentioneering is the framework to explore, grow and optimize your product based on deep analysis of user trajectories. Retentioneering provides systematic and quantitative approach to search for insights, continuous KPI optimization, product improvement and marketing optimization.

    Retentioneering analysis can:
    - Reveal why customers leave your product and reduce churn rate
    - Increase conversion rate and engagement as well as LTV, ARPU, NPS and ROI
    - Identify hidden patterns how customers use your app or website
    - Improve the quality of incoming traffic due to personalization of advertising and increase its conversion to applications
    """,

    author='Retentioneering User Trajectory Analysis Lab',
    author_email='retentioneering@gmail.com',
    url='https://github.com/retentioneering/retentioneering-tools',
    keywords=['ANALYTICS', 'CLICKSTREAM', 'RETENTIONEERING', 'RETENTION', 'GRAPHS', 'TRAJECTORIES', 'PREDICTIVE-ANALYTICS', 'CUSTOMER-SEGMENTATION'],
    install_requires=[
        'pandas>=0.24.2',
        'numpy>=1.16.1',
        'networkx>=2.3',
        'seaborn>=0.9.0',
        'scikit-learn>=0.22.1',
        'altair',
        'vega',
        'pymongo',
        'plotly'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)"

    ],
    python_requires=">=3.6",
    packages=find_packages()
)



# Copyright (C) 2019 Maxim Godzi, Anatoly Zaytsev, Dmitrii Kiselev
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import numpy as np
import pandas as pd
import warnings


def _find_last_min(cnt):
    for i in range(99, 0, -1):
        if _check_local_minimum(i - 1, cnt, neighbours=10):
            break
    return i


def _check_local_minimum(idx, cnt, neighbours=10):
    return cnt[idx] == min(cnt[idx - neighbours: idx + neighbours])


def _find_threshold(time_val):
    cnt, thresh = np.histogram(np.log(time_val), bins=100)  # TODO parametrize + visualization
    idx = _find_last_min(cnt)
    return np.exp(thresh[idx])


def split_sessions(data, minimal_thresh=30, lower_filter=1, upper_filter=99):
    """
    Creates ``session`` column with session rank in given dataset.

    Parameters
    --------
    data: pd.DataFrame
        Clickstream data.
    minimal_thresh: int, optional
        Minimal time difference in seconds between two sessions. If  threshold value of time difference between sessions is lower that ``minimal_thresh``, then sessions are split based on this parameter. Default: ``30``
    lower_filter: float, optional
        Minimal percentile of time distribution between events to be considered as different sessions. For instance, equal to 1 would mean that cutoff time diffrecence between events (time difference between pseudosessions) should not be less than 1st percentile of time distribution between events. Range of possible values: [0, 100]. Default: ``1``
    upper_filter: float, optional
        Same as ``lower_filter`` but as an upper limit. Range of possible values: [0, 100]. Default: ``99``

    Returns
    -------
    Creates session column in dataset.

    Return type
    -------
    pd.DataFrame
    """
    time_col = data.retention.retention_config['event_time_col']
    if 'next_timestamp' not in data.columns:
        data.retention._get_shift()
    time_delta = pd.to_datetime(data['next_timestamp']) - pd.to_datetime(data[time_col])
    time_delta = time_delta.dt.total_seconds()
    time_val = time_delta[time_delta.notnull()].values
    time_val = time_val[(time_val > np.percentile(time_val, lower_filter)) &
                        (time_val < np.percentile(time_val, upper_filter))]
    thresh = _find_threshold(time_val)
    if thresh < minimal_thresh:
        thresh = minimal_thresh
    data['session'] = time_delta > thresh
    data['session'] = data.groupby(data.retention.retention_config['index_col']).session.cumsum()
    data['session'] = data.groupby(data.retention.retention_config['index_col']).session.shift(1).fillna(0)


def _learn_lda(data, **kwargs):
    from sklearn.decomposition import LatentDirichletAllocation
    if hasattr(data.retention, 'datatype') and data.retention.datatype == 'features':
        features = data.copy()
    else:
        if 'ngram_range' not in kwargs:
            kwargs.update({'ngram_range': (1, 2)})
        features = data.retention.extract_features(**kwargs)
    lda_filter = LatentDirichletAllocation.get_params(LatentDirichletAllocation)
    if 'random_state' not in kwargs:
        kwargs.update({'random_state': 0})
    kwargs = {i: j for i, j in kwargs.items() if i in lda_filter}
    lda = LatentDirichletAllocation(**kwargs)
    lda.fit(features)
    mech_desc = pd.DataFrame(lda.components_, columns=features.columns)
    return mech_desc, lda


def _map_mechanic_names(res, main_event_map):
    x = res.loc[:, main_event_map.keys()].copy()
    x = x.rename_axis(main_event_map, axis=1)
    mapper = {}
    for i in range(x.shape[0]):
        name = x.max().argmax()
        idx = x[name].argmax()
        print(name, idx)
        mapper.update({idx: name})
        x = x.drop(idx)
        x = x.drop(name, 1)
        if (x < 1).sum().sum() == x.shape[0] ** 2:
            warnings.warn('Cannot allocate mechanics by given events: {}'.format(' '.join(x.columns)))
            break
    return mapper


def weight_by_mechanics(data, main_event_map, **kwargs):
    """
    Calculates weights of mechanics over ``index_col``.

    Parameters
    --------
    data: pd.DataFrame
        Сlickstream or features data.
    main_event_map: dict
        Mapping of main events into mechanics.
    kwargs: optional
        ``sklearn.decomposition.LatentDirichletAllocation()`` and ``BaseDataset.retention.extract_features()`` parameters.

    Returns
    -------
    Weights of mechanics for each ``index_col`` and mechanics description.
    """
    mech_desc, lda = _learn_lda(data, **kwargs)
    mechanics = _map_mechanic_names(mech_desc, main_event_map)
    res = pd.DataFrame(lda.transform(data)).rename(mechanics, axis=1)
    mech_desc = mech_desc.rename(mechanics)
    setattr(res.retention, 'datatype', 'features')
    return res, mech_desc


def _event_filter_equal(x, y):
    return x != y


def _event_filter_startswith(x, y):
    return ~x.str.startswith(y)


def _event_filter_contains(x, y):
    return ~x.map(lambda z: y in z)
